module Draw (..) where

import Dict exposing (Dict)
import Graphics.Collage exposing (..)
import Color exposing (Color)
import Text
import Types exposing (..)
import Tools


squareSize : Float
squareSize =
  20


colorDict : Dict Int Color
colorDict =
  Dict.fromList
    [ ( 1, Color.blue )
    , ( 2, Color.darkGreen )
    , ( 3, Color.orange )
    , ( 4, Color.darkBlue )
    , ( 5, Color.darkRed )
    , ( 6, Color.lightBlue )
    , ( 7, Color.purple )
    , ( 8, Color.black )
    ]


numberColor : Int -> Color
numberColor n =
  Maybe.withDefault Color.white <| Dict.get n colorDict


drawNumber : Int -> Form
drawNumber n =
  if n > 0 && n <= 9 then
    move ( 0, squareSize / 8 )
      << text
      << Text.bold
      << Text.color (numberColor n)
      << Text.height (0.95 * squareSize)
      << Text.fromString
      << toString
      <| n
  else
    text <| Text.fromString ""


drawButtonUp : Float -> Form
drawButtonUp size =
  group
    [ filled Color.white
        <| polygon
            [ ( size / 2, size / 2 )
            , ( -size / 2, -size / 2 )
            , ( -size / 2, size / 2 )
            ]
    , filled Color.lightCharcoal
        <| polygon
            [ ( -size / 2, -size / 2 )
            , ( size / 2, size / 2 )
            , ( size / 2, -size / 2 )
            ]
    , filled Color.grey
        <| rect (0.8 * size) (0.8 * size)
    ]


drawButtonDown : Float -> Form
drawButtonDown size =
  group
    [ outlined (solid Color.darkGrey)
        <| rect size size
    , filled Color.grey
        <| rect size size
    ]


drawHidden : Form
drawHidden =
  drawButtonUp squareSize


drawExposed : Form
drawExposed =
  drawButtonDown squareSize


drawFlag : Form
drawFlag =
  group
    [ filled Color.black
        <| polygon
            [ ( -squareSize / 6, -squareSize / 3 )
            , ( squareSize / 12, -squareSize / 6 )
            , ( squareSize / 3, -squareSize / 3 )
            ]
    , filled Color.red
        <| polygon
            [ ( squareSize / 12, 0 )
            , ( squareSize / 12, squareSize / 3 )
            , ( -squareSize / 3, squareSize / 8 )
            ]
    , traced defaultLine
        <| segment
            ( squareSize / 12, -squareSize / 6 )
            ( squareSize / 12, squareSize / 3 )
    ]


drawBomb : Form
drawBomb =
  let
    bl =
      solid Color.black

    ls w =
      { bl | width = w }
  in
    group
      [ traced (ls <| squareSize / 12)
          <| segment
              ( -0.45 * squareSize, 0 )
              ( 0.45 * squareSize, 0 )
      , traced (ls <| squareSize / 12)
          <| segment
              ( 0, -0.45 * squareSize )
              ( 0, 0.45 * squareSize )
      , traced (ls <| squareSize / 20)
          <| segment
              ( -0.3 * squareSize, -0.3 * squareSize )
              ( 0.3 * squareSize, 0.3 * squareSize )
      , traced (ls <| squareSize / 20)
          <| segment
              ( 0.3 * squareSize, -0.3 * squareSize )
              ( -0.3 * squareSize, 0.3 * squareSize )
      , gradient
          (Color.radial
            ( -squareSize / 12, squareSize / 10 )
            (squareSize / 20)
            ( -squareSize / 8, squareSize / 12 )
            (0.3 * squareSize)
            [ ( 0, Color.lightGrey ), ( 0.4, Color.black ) ]
          )
          (circle <| 0.3 * squareSize)
      ]


drawCrossedBomb : Form
drawCrossedBomb =
  let
    red =
      solid Color.red

    ls =
      { red | width = squareSize / 12 }
  in
    group
      [ drawBomb
      , traced ls
          <| segment
              ( -squareSize / 3, -squareSize / 3 )
              ( squareSize / 3, squareSize / 3 )
      , traced ls
          <| segment
              ( squareSize / 3, -squareSize / 3 )
              ( -squareSize / 3, squareSize / 3 )
      ]


drawExploded : Form
drawExploded =
  group
    [ filled Color.lightRed
        <| rect squareSize squareSize
    , drawBomb
    ]


drawPressed : Form
drawPressed =
  drawExposed


drawSquare : Bool -> Maybe Square -> Form
drawSquare lost square =
  case square of
    Nothing ->
      drawHidden

    Just { isBomb, isExposed, adjacentBombs, isFlagged, isExploded, isPressed } ->
      if isExploded then
        drawExploded
      else if not isExposed then
        if isFlagged then
          if not isBomb && lost then
            group [ drawExposed, drawCrossedBomb ]
          else
            group [ drawHidden, drawFlag ]
        else if lost && isBomb then
          group [ drawExposed, drawBomb ]
        else if isPressed then
          drawPressed
        else
          drawHidden
      else
        group [ drawExposed, drawNumber adjacentBombs ]


drawBoard : Bool -> Int -> Int -> Board -> Form
drawBoard lost width height board =
  move
    ( (-0.5 * toFloat width) * squareSize
    , (-0.5 * toFloat height) * squareSize
    )
    <| group
    <| List.map
        (\( x, y ) ->
          move
            ( (toFloat x + 0.5) * squareSize
            , (toFloat y + 0.5) * squareSize
            )
            <| drawSquare lost (Dict.get ( x, y ) board)
        )
    <| Tools.product [0..width - 1] [0..height - 1]


drawSmiley : Float -> Smiley -> Form
drawSmiley size smiley =
  let
    body =
      [ filled (Color.rgb 255 255 0) <| circle <| size / 2
      , outlined (solid Color.black) <| circle <| size / 2
      ]

    eyes =
      [ move ( -size / 6, size / 6 )
          <| filled Color.black
          <| circle
          <| size
          / 15
      , move ( size / 6, size / 6 )
          <| filled Color.black
          <| circle
          <| size
          / 15
      ]

    deadEyes =
      [ move ( -size / 6, size / 6 )
          <| traced (solid Color.black)
          <| segment ( -size / 10, -size / 10 ) ( size / 10, size / 10 )
      , move ( -size / 6, size / 6 )
          <| traced (solid Color.black)
          <| segment ( -size / 10, size / 10 ) ( size / 10, -size / 10 )
      , move ( size / 6, size / 6 )
          <| traced (solid Color.black)
          <| segment ( -size / 10, -size / 10 ) ( size / 10, size / 10 )
      , move ( size / 6, size / 6 )
          <| traced (solid Color.black)
          <| segment ( -size / 10, size / 10 ) ( size / 10, -size / 10 )
      ]

    sunGlasses =
      [ move ( -size / 6, size / 8 )
          <| filled Color.black
          <| circle
          <| size
          / 8
      , move ( size / 6, size / 8 )
          <| filled Color.black
          <| circle
          <| size
          / 8
      , traced (solid Color.black)
          <| path
              [ ( -size / 2.5, size / 3 )
              , ( -size / 4, size / 8 )
              , ( size / 4, size / 8 )
              , ( size / 2.5, size / 3 )
              ]
      ]

    normalMouth =
      [ traced (solid Color.black)
          <| path
              [ ( -size / 4, -size / 10 )
              , ( -size / 6, -size / 5 )
              , ( -size / 12, -size / 4 )
              , ( size / 12, -size / 4 )
              , ( size / 6, -size / 5 )
              , ( size / 4, -size / 10 )
              ]
      ]

    deadMouth =
      [ traced (solid Color.black)
          <| path
              [ ( -size / 4, -size / 4 )
              , ( -size / 5, -size / 6 )
              , ( -size / 12, -size / 10 )
              , ( size / 12, -size / 10 )
              , ( size / 5, -size / 6 )
              , ( size / 4, -size / 4 )
              ]
      ]

    oMouth =
      [ move ( 0, -size / 5 )
          <| outlined (solid Color.black)
          <| circle
          <| size
          / 8
      ]
  in
    case smiley of
      SmileyNormal ->
        group <| body ++ eyes ++ normalMouth

      SmileyClicking ->
        group <| body ++ eyes ++ oMouth

      SmileyWon ->
        group <| body ++ sunGlasses ++ normalMouth

      SmileyLost ->
        group <| body ++ deadEyes ++ deadMouth
