module Minesweeper (minesweeperElement, main) where

import Dict exposing (Dict)
import Graphics.Collage exposing (..)
import Graphics.Element exposing (..)
import Graphics.Input exposing (..)
import Random exposing (Generator, Seed)
import Signal exposing (Mailbox)
import Time exposing (Time)
import Text
import Color
import Mouse
import Keyboard
import Char
import Tools
import Draw
import Types exposing (..)
import Generate
import SevenSegment


{-
Notes:
- left/right click
- const Generator
- arcs in Collage
- Module.record.entry does not work
- Module.record/general function calls in record updating syntax
- initialSeed does not depend well on input
- isJust, isNothing
- empty form
- exponentiation with ints and negative exponent
-}


exposeSquare : BoardPos -> Board -> ( Board, Int )
exposeSquare ( x, y ) board =
  case Dict.get ( x, y ) board of
    Nothing ->
      ( board, 0 )

    Just square ->
      if not square.isExposed && not square.isFlagged then
        (\b ->
          if square.adjacentBombs <= 0 then
            List.foldr
              (\pos ( b, n ) ->
                let
                  ( b', n' ) =
                    exposeSquare pos b
                in
                  ( b', n + n' )
              )
              ( b, 1 )
              (List.drop 1
                <| Tools.product [ x, x + 1, x - 1 ] [ y, y + 1, y - 1 ]
              )
          else
            ( b, 1 )
        )
          <| Dict.update
              ( x, y )
              (Maybe.map (\s -> { s | isExposed = True }))
              board
      else
        ( board, 0 )


updateBoardClick : Time -> BoardPos -> Model -> Model
updateBoardClick time pos ({ board, lost, numExposed } as model) =
  if lost || won model then
    model
  else
    case board of
      Nothing ->
        { model | generateNewBoard = Just pos, startTime = time }

      Just b ->
        case Dict.get pos b of
          Nothing ->
            model

          Just square ->
            if square.isFlagged then
              model
            else if square.isBomb then
              { model
                | board =
                    Just
                      <| Dict.insert
                          pos
                          { square | isExploded = True }
                          b
                , lost = True
                , endTime = time
              }
            else
              (\m ->
                let
                  { board } =
                    m
                in
                  if won m then
                    { m
                      | board =
                          Maybe.map
                            (Dict.map
                              (\_ s ->
                                let
                                  { isBomb } =
                                    s
                                in
                                  { s | isFlagged = isBomb }
                              )
                            )
                            board
                      , endTime = time
                    }
                  else
                    m
              )
                <| let
                    ( b', n ) =
                      exposeSquare pos b
                   in
                    { model
                      | board = Just <| b'
                      , numExposed = numExposed + n
                    }


updateBoardFlagClick : BoardPos -> Model -> Model
updateBoardFlagClick pos ({ board, lost, numFlags } as model) =
  if lost || won model then
    model
  else
    case board of
      Nothing ->
        model

      Just b ->
        case Dict.get pos b of
          Nothing ->
            model

          Just ({ isFlagged } as square) ->
            { model
              | board =
                  Just
                    <| Dict.insert
                        pos
                        { square | isFlagged = not isFlagged }
                        b
              , numFlags =
                  numFlags
                    + if isFlagged then
                        -1
                      else
                        1
            }


resetBoard : Model -> Model
resetBoard model =
  { model
    | board = Nothing
    , lost = False
    , numExposed = 0
    , numFlags = 0
    , startTime = -1
    , generateNewBoard = Nothing
  }


uiMailbox : Mailbox Action
uiMailbox =
  Signal.mailbox SmileyButtonPressed


updateAction : Time -> Action -> Model -> Model
updateAction time action model =
  case action of
    SmileyButtonPressed ->
      resetBoard model

    DifficultySelected ( width, height, bombs ) ->
      let
        m =
          resetBoard model
      in
        { m
          | boardWidth = width
          , boardHeight = height
          , numBombs = bombs
        }

    Click pos ->
      case boardCoor pos model of
        Nothing ->
          model

        Just p ->
          updateBoardClick time p model

    FlagClick pos ->
      case boardCoor pos model of
        Nothing ->
          model

        Just p ->
          updateBoardFlagClick p model


generateIfNecessary : ( Model, Seed ) -> ( Model, Seed )
generateIfNecessary ( model, seed ) =
  case model.generateNewBoard of
    Nothing ->
      ( model, seed )

    Just pos ->
      let
        ( board, seed' ) =
          Random.generate
            (Generate.boardGenerator
              pos
              model.boardWidth
              model.boardHeight
              model.numBombs
            )
            seed

        model' =
          { model | generateNewBoard = Nothing }
      in
        case board of
          Nothing ->
            ( model', seed' )

          Just b ->
            let
              ( b', n ) =
                exposeSquare pos b
            in
              ( { model'
                  | board = Just b'
                  , numExposed = n
                }
              , seed'
              )


borderWidth : Int
borderWidth =
  7


uiBorderWidth : Int
uiBorderWidth =
  5


uiInnerHeight : Int
uiInnerHeight =
  30


uiHeight : Int
uiHeight =
  uiInnerHeight + 2 * uiBorderWidth


boardCoor : ( Int, Int ) -> Model -> Maybe BoardPos
boardCoor ( x, y ) model =
  let
    xPos =
      floor <| toFloat (x - borderWidth) / Draw.squareSize

    yPos =
      floor
        <| (toFloat model.boardHeight
              * Draw.squareSize
              + toFloat uiHeight
              + toFloat borderWidth
              * 2
              - toFloat y
           )
        / Draw.squareSize
  in
    if
      0
        <= xPos
        && xPos
        < model.boardWidth
        && 0
        <= yPos
        && yPos
        < model.boardHeight
    then
      Just ( xPos, yPos )
    else
      Nothing


smileyButton : Smiley -> Element
smileyButton smiley =
  let
    width =
      uiInnerHeight

    size =
      toFloat width

    pic =
      Draw.drawSmiley (size * 0.6) smiley
  in
    customButton
      (Signal.message uiMailbox.address SmileyButtonPressed)
      (collage width width
        <| [ Draw.drawButtonUp size, pic ]
      )
      (collage width width
        <| [ Draw.drawButtonUp size, pic ]
      )
      (collage width width
        <| [ Draw.drawButtonDown size, pic ]
      )


view : Time -> Maybe BoardPos -> Bool -> Model -> Element
view time pos isDown model =
  let
    boardEl =
      collage
        (ceiling <| toFloat model.boardWidth * Draw.squareSize)
        (ceiling <| toFloat model.boardHeight * Draw.squareSize)
        <| flip (::) []
        <| Draw.drawBoard model.lost model.boardWidth model.boardHeight
        <| (if model.lost || won model || not isDown then
              identity
            else
              Dict.update
                (Maybe.withDefault ( -1, -1 ) pos)
                (\s ->
                  case s of
                    Nothing ->
                      let
                        i =
                          Generate.initSquare
                      in
                        Just { i | isPressed = True }

                    Just square ->
                      Just { square | isPressed = True }
                )
           )
        <| Maybe.withDefault Dict.empty model.board

    smileyButtonEl =
      smileyButton
        <| if won model then
            SmileyWon
           else if model.lost then
            SmileyLost
           else if Tools.isJust pos && isDown then
            SmileyClicking
           else
            SmileyNormal

    timeEl =
      SevenSegment.sevenSegment uiInnerHeight 3
        <| if Tools.isJust model.board then
            ceiling
              <| Time.inSeconds
              <| if model.lost || won model then
                  model.endTime - model.startTime
                 else
                  time - model.startTime
           else
            0

    numBombsEl =
      SevenSegment.sevenSegment uiInnerHeight 3
        <| model.numBombs
        - model.numFlags

    difficultyDropDown =
      dropDown
        (Signal.message uiMailbox.address << DifficultySelected)
        [ ( "Beginner", ( 9, 9, 10 ) )
        , ( "Intermediate", ( 16, 16, 40 ) )
        , ( "Expert", ( 30, 16, 99 ) )
        ]

    difficultyEl =
      container
        (widthOf boardEl)
        (heightOf difficultyDropDown + borderWidth)
        bottomLeft
        (flow
          right
          [ leftAligned <| Text.fromString "Difficulty:  "
          , difficultyDropDown
          ]
        )

    uiInnerCont =
      container (widthOf boardEl - 2 * uiBorderWidth) uiInnerHeight

    uiEl =
      color Color.darkGrey
        <| container (widthOf boardEl) uiHeight middle
        <| layers
            [ uiInnerCont middle smileyButtonEl
            , uiInnerCont midRight timeEl
            , uiInnerCont midLeft numBombsEl
            ]

    gameEl =
      flow
        down
        [ uiEl
        , spacer 1 borderWidth
        , boardEl
        , difficultyEl
        , spacer 1 borderWidth
        , leftAligned <| Text.fromString "Hold F to set flags."
        ]
  in
    color Color.lightCharcoal
      <| container
          (widthOf gameEl + 2 * borderWidth)
          (heightOf gameEl + 2 * borderWidth)
          middle
          gameEl


minesweeperElement : Signal Element
minesweeperElement =
  let
    flagKeySig =
      Keyboard.isDown <| Char.toCode 'F'

    modelSig =
      (Signal.map fst
        <| Signal.foldp
            (\( time, a ) ( m, s ) ->
              let
                seed =
                  case s of
                    Nothing ->
                      Random.initialSeed <| floor time

                    Just s' ->
                      s'

                ( m', seed' ) =
                  generateIfNecessary
                    ( updateAction time a m, seed )
              in
                ( m', Just seed' )
            )
            ( Generate.initModel, Nothing )
            (Time.timestamp
              <| Signal.merge
                  uiMailbox.signal
                  (Signal.map
                    (\( down, pos ) ->
                      if down then
                        FlagClick pos
                      else
                        Click pos
                    )
                    <| Signal.sampleOn Mouse.clicks
                    <| Signal.map2
                        (,)
                        flagKeySig
                        Mouse.position
                  )
            )
      )
  in
    Signal.map4
      view
      (Time.every Time.second)
      (Signal.map2 boardCoor Mouse.position modelSig)
      (Signal.map2 (\m k -> m && not k) Mouse.isDown flagKeySig)
      modelSig


main : Signal Element
main =
  minesweeperElement
